/*
 * Copyright (C) 2001 LAAS/CNRS 
 *
 * Anthony Mallet - Wed Jul  4 2001
 *
 * $Log: p3d.h,v $
 * Revision 1.4  2007/11/22 10:06:25  adegroot
 * Add a workaround in case where the robot is not fully in the dtm.
 *
 * If we can't compute the configuration of back axle, create a fake configuration
 * based on the real position of the robot.
 *
 * Revision 1.3  2007/11/22 09:52:46  adegroot
 * nIter must be an int, not in double.
 *
 * Revision 1.2  2007/01/08 19:55:17  sjoyeux
 * * sync with the DEM tag (was something else before)
 * * update to OpenRobot's T3Ds
 *
 * Revision 1.5.2.11  2006/04/18 17:05:00  tpeynot
 * DIF_MAP structure declaration moved to separate file difmap.h and structure changed: now contains a cell tab with more info than just the difficulty
 *
 * Revision 1.5.2.10  2006/03/18 21:43:53  tpeynot
 * difmap->difficulty is now a dynamic array
 *
 * Revision 1.5.2.9  2006/02/01 13:10:53  tpeynot
 * MAX p3dmap lines and cols increased and sigmas added to config struct
 *
 * Revision 1.5.2.8  2005/11/24 15:17:05  tpeynot
 * sigma_z, nb_points and confidence added to struct + max size increased to 400
 *
 * Revision 1.5.2.7  2005/07/13 15:58:52  tpeynot
 * definition of a DIF_MAP (Difficulty map) added
 *
 * Revision 1.5.2.6  2005/04/21 12:40:27  tpeynot
 * just a more precise comment
 *
 * Revision 1.5.2.5  2005/04/18 10:00:35  tpeynot
 * Add flag ATRV_JUNIOR together with ATRV
 *
 * Revision 1.5.2.4  2005/04/13 23:21:21  tpeynot
 * Add elements for body-ground collision test
 *
 * Revision 1.5.2.3  2005/03/29 09:37:12  tpeynot
 * the p3dMap->mainToOrigin is now a T3D*
 *
 * Revision 1.5.2.2  2005/03/21 23:07:39  tpeynot
 * Add prototypes of functions create_p3dmap and write_p3dmap
 *
 * Revision 1.5.2.1  2005/03/18 23:47:44  tpeynot
 * Declaration of struct P3D_MAP in p3d.h
 * and take it into account in p3dPlan prototype in libp3d.h
 *
 * Revision 1.5  2003/09/12 15:21:10  mallet
 * correctly handle arcs that cross goalTol limit
 *
 * Revision 1.4  2002/05/22 09:07:02  mallet
 * sync for dala
 *
 * Revision 1.3  2001/07/14 11:38:48  mallet
 * premiere version ok
 *
 * Revision 1.2  2001/07/12 17:38:44  mallet
 * sync
 *
 * Revision 1.1  2001/07/11 18:06:14  mallet
 * headers pour p3d
 *
 */
#ifndef P3D_H
#define P3D_H

#include <t3d/t3d.h>

#ifdef __cplusplus
extern "C" {
#endif
/*
 * Robot-specific structures --------------------------------------------
 *
 * This should be the only robot-dependant structs, and one should
 * provide specific function to compute generic values based on these
 * stuctures.
 *
 * XXX We do not bother whith pseudo-generic stuff until several robots
 * must actually be considered.
 */

#define HAS_BODY_SUPPORT defined(ATRV) || defined(ATRV_JUNIOR) || defined(TROOPER) || defined(RMP400) || defined(RMP440)

/* Lama */
typedef struct P3DRobotGeometry {
   double len[3];	/* wheel lengths */
   double rad[2];	/* wheel radiuses */
   double arm[2];	/* front and back arm lengths */

#if HAS_BODY_SUPPORT
  /* params de bas de caisse */
  double bodyLength;  /* Length and width of ATRV's under body */
  double bodyWidth;
  double zBody;      /* dz between Robot's center and under body */
  /* dz entre centre du robot et bas de caisse */
  double groundClear;  /* ground clearance (garde au sol) */
#endif

  /* Limits of angle : 
   * xLimits[0] is the limit between the easy and the medium configuration angle
   * xLimits[1] is the limit betwwen the medium and hard configuration angle
   * xLimits[2] is the limit between hard and impossible angle
   */

  double rollLimits[3];
#define rollMax rollLimits[2]
  double pitchLimits[3];
#define pitchMax pitchLimits[2]
  double alphaLimits[3];
#define alphaMax alphaLimits[2]
  double betaLimits[3];
#define betaMax betaLimits[2]

   double unkMax;	/* max percentage of unkown terrain */
} P3DRobotGeometry; 

#define		EASY	(4.0)
#define		MEDIUM	(8.0)
#define		HARD	(16.0)

/* Lama */
typedef struct P3DRobotConfig {
   double q;		/* quality in [0;1], <0 indicates error */

   double x,y,z,theta;	/* Robot's position wrt the map */
  
  double dzBodyGround;  /* min. height between robot body bottom and ground */

   double roll, pitch;	/* of middle axle */
   double alpha1, alpha2;/* front and back axle roll */
   double beta3;	/* angle between back and front arm */

   double pUnk;		/* percentage of unknown terrain */

   double sigmaPitch;   /* sigma associated to config. placement */
   double sigmaRoll;
  double sigmaz;
} P3DRobotConfig; 


/*
 * Internal robot geometry -------------------------------------------
 *
 * XXX Elements are generic wheel profiles, but it is still
 * Lama-dependant since P3DRobot only consider left and right wheels.
 * Who cares? 
 */

/* The same structure is used to compute a bitmap of Dala's body */

typedef struct P3DElement {
   int w, h;		/* size of height map */
   float x,y;		/* reference point inside bitmap (meters) */
   float scale;		/* size of pixels of the bitmap (meters/pixel) */

#define P3D_UNDEF_ELEM	(32768.0) /* why not */

   float *height;	/* heights */
   int nElem;		/* number of valid heights */
} P3DElement;

typedef struct P3DRobot {
	double damping;	/* Damping factor for axle placement */
	int nIter;	/* Max number of iterations for axle placement */

	double rollMin, rollStep;

	double bodyPitchMin, bodyPitchStep;   /* This is just used for ATRV  */
	double bodyRollMin, bodyRollStep;

	int nProfiles;	/* number of wheel bitmaps */
	P3DElement **lWheel;	/* array of bitmaps indexed by roll angle */
	P3DElement **rWheel;	/* array of bitmaps indexed by roll angle */

#if HAS_BODY_SUPPORT
	int nBodyProfiles;  /* number of body bitmaps */
	int nBodyRoll, nBodyPitch;  /* number of body maps indexed by roll, pitch */
	P3DElement ***body;	/* array of bitmaps indexed by roll AND pitch angles */
#endif

	double vMin, vMax; /* Low and High bound for robot speed */
	double wMax; /* High bound for robot rotation speed */
	double accel, decel; /* Acceleration and decelleration in m/s-2 */

} P3DRobot;


/* Arcs ----------------------------------------------------------------- */

typedef struct P3DNode {
   double x,y,theta;	/* horizontal position relative to the robot */
  
   P3DRobotConfig conf;	/* robot configuration */
   P3DRobotConfig *prev;/* previous configuration */

#define P3D_UNDEF_COST	(-16384.0)	/* (should be negative) */
#define P3D_IMPOSSIBLE_COST (16384.0)  /* it is the cost in an impossible situation */

   double cost;		/* to reach this node from previous one */
   double remain;	/* to reach goal from this node */

   int number;		/* node number */
   int isLocalMin;	/* true if distance to goal is a local minimum */
   struct P3DArc *parent;/* arc of this node */
} P3DNode;


#define P3D_MIN_DISTANCE	(1.0)	/* We take one meter more than the stop distance */

typedef struct P3DArc {
   double r;		/* arc curvature */
   double ds;		/* nodes spacing */
   int depth;		/* depth (not used at the moment) */

   int n;		/* number of nodes */
   P3DNode *nodes;	/* array of configurations */

   int current;		/* last computed node */
   int first_unknow; /* First node in unknow world */
   double cost;		/* to reach "current" */
   double cost_balanced; /* Only used for P3D_FULL algo */

   int nArcs;		/* number of arcs starting after last node */
   struct P3DArc *arcs;	/* array of sucessors */

   int number;		/* arc number */
   struct P3DArc *parent;/* arc parent (NULL for toplevel) */

   int is_elected; 
} P3DArc;

/* 
 * The following structure are quite similar to P3DNode and P3dArc
 * The only difference is that they are linear in memory and so it can be
 * used in poster
 */

typedef struct P3DNodeDump {
	double x;
	double y;
	double z;
	double yaw;
	double pitch;
	double roll;
	double pUnk;
	double cost;
	double remain;
} P3DNodeDump;

#define MAX_NUMBER_NODE 20

typedef struct P3DArcDump {
	int nArcsTotal;		/* only filled for the first P3DArcDump */
	double r;
	double cost;
	double cost_balanced;
	int depth;
	int nArcs;
	int nNodes;
	int current;
	int first_unknow;
	P3DNodeDump nodes[MAX_NUMBER_NODE];
	int is_elected;
} P3DArcDump;


/* Planner data ------------------------------------------------------ */

typedef struct P3DPlanner {
   P3DRobotGeometry *geom;	/* human readable geometry */
   P3DRobot *robot;		/* internal geometric description */

   int nArcs;			/* number of arcs considered */
   P3DArc *arcs;		/* array of arcs */
   double goalTol;		/* distance under which the goal is
				 * reached */
   double dMin;			/* Minimal size of each arc */
} P3DPlanner;


/*-----------------------------------------------------------------------*/

/* Planner Resultat data ----------------------------------------------- */

typedef struct P3DPlanSolution {
	int is_valid;

	double curv;		
	double d;
	double cost; 

	double v;			/* speed in m/s */
	double w;			/* angular speed in rad/s */
	double alpha;		/* wheel angle in rad */

	int depth;
	int _maxdepth;		/* Internal, use for bestArc allocation */
	P3DArc* bestArc;

	double average_time; /* Average time to process p3d */
	unsigned int ncalls; /* number of call of p3d */

} P3DPlanSolution;

/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* Map structure specific to p3d, but equivalent to a DEM_P3D_POSTER */

/*  so as to use a Calife struct, not a genom module struct, such as DEM_P3D_POSTER */

/* These params should be at least the same as DEM_MAX_LINES and DEM_MAX_COLUMNS in demStruct.h (or higher) => should be checked when copying the demP3dPoster in a P3D_MAP in the p3d genom Module */
#define	P3D_MAP_MAX_LINES	1000
#define	P3D_MAP_MAX_COLUMNS	1000	/* Both sizes must be even, so that the
				   structure is aligned wrt. doubles */

typedef enum p3dMapCellState {P3D_MAP_CELL_EMPTY, P3D_MAP_CELL_FILLED} p3dMapCellState;


typedef struct P3D_MAP_CELL {

  p3dMapCellState  state;  /* State of the cell: EMPTY or FULL */
  float            z;     /* average elevation on this cell */
  float            sigma_z;   /* standard deviation */
  int              nb_points;  /* Number of points fused in this dem cell */
  float            confidence;  /* Confidence on z and luminance info in the dem */

} P3D_MAP_CELL;



typedef struct P3D_MAP {
  int		nbLines;
  int		nbCols;		/* Nb of lines and columns of the dem */

  float		xOrigin;
  float		yOrigin;	/* Coordinates of the topleft corner of the
				   (0,0) pixel in the global frame */

  T3D*          mainToOrigin;   /* Position of the (0,0) point of the map wrt the global frame */

  float		scale;		/* Size of a cell in meters */
  /* we should have xScale = yScale (=scale): to be checked in the demP3dPoster */

  float		zOrigin;	/* Elevation of the '0' value */
  float		zScale;		/* Scale to store the elevation */

  P3D_MAP_CELL		cells[P3D_MAP_MAX_LINES][P3D_MAP_MAX_COLUMNS];
} P3D_MAP;

enum P3DWay {
	P3D_ASTAR, /* A* implementation */
	P3D_FULL  /* Full exploration with ponderation */
};

#ifdef __cplusplus
}
#endif

#endif /* P3D_H */
